import struct
import mapping

class NotImplementedException(Exception):
    pass

class ConnectionError(Exception):
    pass

class Transport(object):
    def __init__(self, device, *args, **kwargs):
        self.device = device
        self.session_depth = 0
        self._open()
    
    def _open(self):
        raise NotImplementedException("Not implemented")
    
    def _close(self):
        raise NotImplementedException("Not implemented")
    
    def _write(self, msg, protobuf_msg):
        raise NotImplementedException("Not implemented")
    
    def _read(self):
        raise NotImplementedException("Not implemented")
    
    def _session_begin(self):
        pass
    
    def _session_end(self):
        pass
    
    def ready_to_read(self):
        """
        Returns True if there is data to be read from the transport.  Otherwise, False.
        """
        raise NotImplementedException("Not implemented")
    
    def session_begin(self):
        """
        Apply a lock to the device in order to preform synchronous multistep "conversations" with the device.  For example, before entering the transaction signing workflow, one begins a session.  After the transaction is complete, the session may be ended.
        """
        if self.session_depth == 0:
            self._session_begin()
        self.session_depth += 1
    
    def session_end(self):
        """
        End a session.  Se session_begin for an in depth description of TREZOR sessions.
        """
        self.session_depth -= 1
        self.session_depth = max(0, self.session_depth)
        if self.session_depth == 0:
            self._session_end()
        
    def close(self):
        """
        Close the connection to the physical device or file descriptor represented by the Transport.
        """
        self._close()
        
    def write(self, msg):
        """
        Write mesage to tansport.  msg should be a member of a valid `protobuf class <https://developers.google.com/protocol-buffers/docs/pythontutorial>`_ with a SerializeToString() method.
        """
        ser = msg.SerializeToString()
        header = struct.pack(">HL", mapping.get_type(msg), len(ser))
        self._write("##%s%s" % (header, ser), msg)

    def read(self):
        """
        If there is data available to be read from the transport, reads the data and tries to parse it as a protobuf message.  If the parsing succeeds, return a protobuf object.
        Otherwise, returns None.
        """
        if not self.ready_to_read():
            return None

        data = self._read()
        if data == None:
            return None
        
        return self._parse_message(data)
        
    def read_blocking(self):
        """
        Same as read, except blocks untill data is available to be read.
        """
        while True:
            data = self._read()
            if data != None:
                break
        
        return self._parse_message(data)

    def _parse_message(self, data):
        (msg_type, data) = data
        if msg_type == 'protobuf':
            return data
        else:
            inst = mapping.get_class(msg_type)()
            inst.ParseFromString(data)
            return inst
    
    def _read_headers(self, read_f):
        # Try to read headers until some sane value are detected
        is_ok = False
        while not is_ok:

            # Align cursor to the beginning of the header ("##")
            c = read_f.read(1)
            i = 0
            while c != '#':
                i += 1
                if i >= 64:
                    # timeout
                    raise Exception("Timed out while waiting for the magic character")
                #print "Aligning to magic characters"
                c = read_f.read(1)

            if read_f.read(1) != "#":
                # Second character must be # to be valid header
                raise Exception("Second magic character is broken")

            # Now we're most likely on the beginning of the header
            try:
                headerlen = struct.calcsize(">HL")
                (msg_type, datalen) = struct.unpack(">HL", read_f.read(headerlen))
                break
            except:
                raise Exception("Cannot parse header length")
       
        return (msg_type, datalen)



'''PipeTransport implements fake wire transport over local named pipe.
Use this transport for talking with trezor simulator.'''

import os
from select import select
from transport import Transport

class PipeTransport(Transport):
    def __init__(self, device, is_device, *args, **kwargs):
        self.is_device = is_device # Set True if act as device
        
        super(PipeTransport, self).__init__(device, *args, **kwargs)        
        
    def _open(self):
        if self.is_device:
            self.filename_read = self.device+'.to'
            self.filename_write = self.device+'.from'
             
            os.mkfifo(self.filename_read, 0600)
            os.mkfifo(self.filename_write, 0600)
        else:
            self.filename_read = self.device+'.from'
            self.filename_write = self.device+'.to'

            if not os.path.exists(self.filename_write):
                raise Exception("Not connected")
            
        self.write_fd = os.open(self.filename_write, os.O_RDWR)#|os.O_NONBLOCK)
        self.write_f = os.fdopen(self.write_fd, 'w+')
    
        self.read_fd = os.open(self.filename_read, os.O_RDWR)#|os.O_NONBLOCK)
        self.read_f = os.fdopen(self.read_fd, 'rb')
        
    def _close(self):
        self.read_f.close()
        self.write_f.close()
        if self.is_device:
            os.unlink(self.filename_read)
            os.unlink(self.filename_write)

    def ready_to_read(self):
        rlist, _, _ = select([self.read_f], [], [], 0)
        return len(rlist) > 0
    
    def _write(self, msg, protobuf_msg):
        try:
            self.write_f.write(msg)
            self.write_f.flush()
        except OSError:
            print "Error while writing to socket"
            raise
        
    def _read(self):
        try:
            (msg_type, datalen) = self._read_headers(self.read_f)
            return (msg_type, self.read_f.read(datalen))
        except IOError:
            print "Failed to read from device"
            raise



import binascii
import urllib2
import json
from decimal import Decimal
try:
    raise Exception() # remove this line to enable caching
    from filecache import filecache, DAY
except:
    def filecache(x):
        def _inner(y):
            return y
        return _inner
    DAY = None

import types_pb2 as proto_types

def op_push_data(data):
    l = len(data)
    if l < 0x4C:
        return chr(l) + data
    elif i < 0xFF:
        return '\x4C' + chr(l) + data
    elif i < 0xFFFF:
        return '\x4D' + struct.pack("<H", i) + data
    else:
        return '\x4E' + struct.pack("<I", i) + data

def opcode_serialize(opcode):
    mapping = {
        'OP_TRUE'                : '\x51',
        'OP_RETURN'              : '\x6A',
        'OP_DUP'                 : '\x76',
        'OP_EQUAL'               : '\x87',
        'OP_EQUALVERIFY'         : '\x88',
        'OP_RIPEMD160'           : '\xA6',
        'OP_SHA1'                : '\xA7',
        'OP_SHA256'              : '\xA8',
        'OP_HASH160'             : '\xA9',
        'OP_HASH256'             : '\xAA',
        'OP_CHECKSIG'            : '\xAC',
        'OP_CHECKSIGVERIFY'      : '\xAD',
        'OP_CHECKMULTISIG'       : '\xAE',
        'OP_CHECKMULTISIGVERIFY' : '\xAF',
    }
    # check if it is known opcode
    if mapping.has_key(opcode):
        return mapping[opcode]
    # it's probably hex data
    try:
        x = binascii.unhexlify(opcode)
        return op_push_data(x)
    except:
        raise Exception('Unknown script opcode: %s' % opcode)

def insight_tx(url):
    try:
        f = urllib2.urlopen(url)
    except:
        raise Exception('URL error: %s' % url)
    data = json.load(f)

    t = proto_types.TransactionType()
    t.version = data['version']
    t.lock_time = data['locktime']

    for vin in data['vin']:
        i = t.inputs.add()
        if 'coinbase' in vin.keys():
            i.prev_hash = "\0"*32
            i.prev_index = 0xffffffff # signed int -1
            i.script_sig = binascii.unhexlify(vin['coinbase'])
            i.sequence = vin['sequence']

        else:
            i.prev_hash = binascii.unhexlify(vin['txid'])
            i.prev_index = vin['vout']
            asm = vin['scriptSig']['asm'].split(' ')
            asm = [ opcode_serialize(x) for x in asm ]
            i.script_sig = ''.join(asm)
            i.sequence = vin['sequence']

    for vout in data['vout']:
        o = t.bin_outputs.add()
        o.amount = int(Decimal(vout['value']) * 100000000)
        asm = vout['scriptPubKey']['asm'].split(' ')
        asm = [ opcode_serialize(x) for x in asm ]
        o.script_pubkey = ''.join(asm)

    return t

class TXAPIBitcoin(object):

    @filecache(DAY)
    def get_tx(self, txhash):
        url = 'https://insight.bitpay.com/api/tx/%s' % txhash
        return insight_tx(url)

class TXAPITestnet(object):

    @filecache(DAY)
    def get_tx(self, txhash):
        url = 'https://test-insight.bitpay.com/api/tx/%s' % txhash
        return insight_tx(url)
